open Printf

type t = {
  filename: string;
  offset: int;
  line: int;
  bol: int
}

let of_lexing p = {
  filename = p.Lexing.pos_fname;
  offset = p.Lexing.pos_cnum;
  line = p.Lexing.pos_lnum;
  bol = p.Lexing.pos_bol;
}

type range = {
  first: t;
  last: t
}

type 'a loc = {
  range: range;
  v: 'a;
}

let of_lexbuf p = {
  first = of_lexing p.Lexing.lex_start_p;
  last = of_lexing p.Lexing.lex_curr_p;
}

let dummy = {
  filename = "";
  offset = -1;
  line = -1;
  bol = -1
}

let dummy_range = {
  first = dummy;
  last = dummy;
}

let is_dummy pos = pos.offset = -1

let print chan pos =
  if not (is_dummy pos) then
    if pos.filename = "" then
      fprintf chan "Line %d, character %d:\n"
        pos.line (pos.offset - pos.bol)
    else
      fprintf chan "File %s, line %d, character %d:\n"
        (Filename.quote pos.filename)
        pos.line (pos.offset - pos.bol)

let print_range chan range =
  if range.first.filename <> range.last.filename then
    raise (Invalid_argument "Pos.print");
  if not (is_dummy range.first) then
    if range.first.filename = "" then
      if range.first.line = range.last.line then
        if range.first.offset = range.last.offset then
          fprintf chan "Line %d, character %d:\n"
            range.first.line (range.first.offset - range.first.bol)
        else
          fprintf chan "Line %d, characters %d-%d:\n"
            range.first.line (range.first.offset - range.first.bol)
            (range.last.offset - range.last.bol)
      else
        fprintf chan
          "Line %d, character %d to line %d, character %d:\n"
          range.first.line (range.first.offset - range.first.bol)
          range.last.line (range.last.offset - range.last.bol)
    else
      if not (is_dummy range.first) then
        if range.first.line = range.last.line then
          if range.first.offset = range.last.offset then
            fprintf chan "File %s, line %d, character %d:\n"
              (Filename.quote range.first.filename)
              range.first.line (range.first.offset - range.first.bol)
          else
            fprintf chan "File %s, line %d, characters %d-%d:\n"
              (Filename.quote range.first.filename)
              range.first.line (range.first.offset - range.first.bol)
              (range.last.offset - range.last.bol)
        else
          fprintf chan
            "File %s, line %d, character %d to line %d, character %d:\n"
            (Filename.quote range.first.filename)
            range.first.line (range.first.offset - range.first.bol)
            range.last.line (range.last.offset - range.last.bol)

